#ifndef __BACKPORT_LINUX_PCI_TO_2_6_19__
#define __BACKPORT_LINUX_PCI_TO_2_6_19__

#include_next <linux/pci.h>

#include <linux/delay.h>

/**
 * PCI_VDEVICE - macro used to describe a specific pci device in short form
 * @vend: the vendor name
 * @dev: the 16 bit PCI Device ID
 *
 * This macro is used to create a struct pci_device_id that matches a
 * specific PCI device.  The subvendor, and subdevice fields will be set
 * to PCI_ANY_ID. The macro allows the next field to follow as the device
 * private data.
 */

#define PCI_VDEVICE(vendor, device)            \
	PCI_VENDOR_ID_##vendor, (device),       \
PCI_ANY_ID, PCI_ANY_ID, 0, 0

#define PCI_VPD_LRDT                    0x80    /* Large Resource Data Type */
#define PCI_VPD_LRDT_ID(x)              (x | PCI_VPD_LRDT)
#define PCI_VPD_LRDT_ID_STRING          PCI_VPD_LRDT_ID(PCI_VPD_LTIN_ID_STRING)
#define PCI_VPD_LRDT_RO_DATA            PCI_VPD_LRDT_ID(PCI_VPD_LTIN_RO_DATA)
#define PCI_VPD_LRDT_TAG_SIZE           3
#define PCI_VPD_INFO_FLD_HDR_SIZE       3
#define PCI_VPD_LTIN_ID_STRING          0x02    /* Identifier String */
#define PCI_VPD_LTIN_RO_DATA            0x10    /* Read-Only Data */
#define PCI_VPD_SRDT_LEN_MASK           0x07
#define PCI_VPD_STIN_END                0x78    /* End */
#define PCI_VPD_SRDT_END                PCI_VPD_STIN_END
#define PCI_VPD_SRDT_TAG_SIZE           1
#define PCI_VPD_MAX_POLL        	40
#define PCI_VPD_LEN             	(1 + PCI_VPD_ADDR_MASK)

/*
 * pci_read_vpd - used by eeprom_rd_pyhs
 */
static inline ssize_t pci_read_vpd(struct pci_dev *dev,
		loff_t pos, size_t count, void *arg)
{
	loff_t  end = pos + count;
	u8      *buf = arg;
	u16     v16;
	u32     v32;
	int     attempts = PCI_VPD_MAX_POLL;
	int     ret;

	static struct cache {
		struct pci_dev *dev;
		u8      cap;
	} cache = {0, 0};

	if (pos < 0 || pos > PCI_VPD_LEN || end > PCI_VPD_LEN)
		return -EINVAL;

	if (cache.dev != dev) {
		cache.cap = pci_find_capability(dev, PCI_CAP_ID_VPD);
		cache.dev = dev;
	}

	while (pos < end) {
		unsigned int i, skip;

		ret = pci_write_config_word(dev, cache.cap + PCI_VPD_ADDR,
				pos & ~3);
		if (ret < 0)
			return ret;

		attempts = PCI_VPD_MAX_POLL;
		do {
			udelay(10);
			pci_read_config_word(dev, cache.cap + PCI_VPD_ADDR,
						 &v16);
		} while (!(v16 & PCI_VPD_ADDR_F) && --attempts);

		if (attempts <= 0)
			return -ETIMEDOUT;

		ret = pci_read_config_dword(dev, cache.cap + PCI_VPD_DATA,
						 &v32);
		if (ret < 0)
			break;

		skip = pos & 3;
		for (i = 0;  i < sizeof(u32); ++i) {
			if (i >= skip) {
				*buf++ = v32;
				if (++pos == end)
					break;
			}
			v32 >>= 8;
		}
	}

	return count;
}

static inline ssize_t pci_write_vpd(struct pci_dev *dev,
		loff_t pos, size_t count, const void *arg)
{
	const u8        *buf = arg;
	loff_t          end = pos + count;
	int             ret = 0;
	u16             v16;
	u32             v32;
	int             attempts = PCI_VPD_MAX_POLL;

	static struct cache {
		struct pci_dev *dev;
		u8      cap;
	} cache = {0, 0};

	if (pos < 0 || (pos & 3) || (count & 3) || end > PCI_VPD_LEN)
		return -EINVAL;

	if (cache.dev != dev) {
		cache.cap = pci_find_capability(dev, PCI_CAP_ID_VPD);
		cache.dev = dev;
	}

	attempts = PCI_VPD_MAX_POLL;
	do {
		udelay(10);
		pci_read_config_word(dev, cache.cap + PCI_VPD_ADDR, &v16);
	} while (!(v16 & PCI_VPD_ADDR_F) && --attempts);

	if (attempts <= 0)
		return -ETIMEDOUT;

	while (pos < end) {
		v32  = *buf++ <<  0;
		v32 |= *buf++ <<  8;
		v32 |= *buf++ << 16;
		v32 |= *buf++ << 24;

		ret = pci_write_config_dword(dev, cache.cap + PCI_VPD_DATA,
						 v32);
		if (ret < 0)
			return ret;
		ret = pci_write_config_word(dev, cache.cap + PCI_VPD_ADDR,
				pos | PCI_VPD_ADDR_F);
		if (ret < 0)
			return ret;

		attempts = PCI_VPD_MAX_POLL;
		do {
			udelay(10);
			pci_read_config_word(dev, cache.cap + PCI_VPD_ADDR,
						 &v16);
		} while (!(v16 & PCI_VPD_ADDR_F) && --attempts);

		if (attempts <= 0)
			return -ETIMEDOUT;

		pos += sizeof(u32);
	}
	return count;
}

static inline u32 pci_pcie_cap(struct pci_dev *pci)
{
	return pci_find_capability(pci, PCI_CAP_ID_EXP);
}
#endif
